from __future__ import annotations

from argparse import ArgumentParser, Namespace
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import ClassVar, TextIO

from .._colors import Colors
from .._owners import CodeOwners


@dataclass
class Command:
    name: ClassVar[str]
    args: Namespace
    stdout: TextIO

    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        parser.add_argument(
            '--no-colors', action='store_true',
            help='disable colored output.',
        )
        parser.add_argument(
            '--root', type=Path, default=Path(),
            help='the root directory of the project.',
        )

    def run(self) -> int:
        raise NotImplementedError

    def print(self, *args: str, end: str = '\n', sep: str = ' ') -> None:
        print(*args, file=self.stdout, end=end, sep=sep)

    @cached_property
    def colors(self) -> Colors:
        return Colors(disabled=self.args.no_colors)

    @cached_property
    def code_owners(self) -> CodeOwners:
        return CodeOwners(root=self.args.root.absolute())



from __future__ import annotations

from argparse import ArgumentParser

from ._base import Command


class OwnedBy(Command):
    """Show paths owned by the given user or group.
    """

    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        Command.init_parser(parser)
        parser.add_argument('owners', nargs='+')

    def run(self) -> int:
        owners = set(self.args.owners)
        root = self.code_owners.root
        for rule in self.code_owners.rules:
            if not owners & set(rule.owners):
                continue
            for path in rule.paths:
                path = path.relative_to(root)
                self.print(str(path))
        return 0



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Iterator

from ._rule import Rule
from ._section import Section


@dataclass(frozen=True)
class CodeOwners:
    root: Path

    @cached_property
    def file_path(self) -> Path:
        paths = [
            self.root / 'CODEOWNERS',
            self.root / 'docs' / 'CODEOWNERS',
            self.root / '.gitlab' / 'CODEOWNERS',
        ]
        for path in paths:
            if path.exists():
                return path
        raise FileNotFoundError('cannot find CODEOWNERS')

    @cached_property
    def sections(self) -> tuple[Section, ...]:
        sections: list[Section] = []
        section_name = ''
        rules: list[Rule] = []
        with self.file_path.open('r', encoding='utf8') as stream:
            for line in stream:
                line = line.strip()

                # comment line
                if not line or line.startswith('#'):
                    continue
                is_section = line.startswith(('^[', '['))

                # rule line
                if not is_section:
                    rules.append(Rule(root=self.root, raw=line))
                    continue

                # section line
                if rules:
                    sections.append(Section(
                        root=self.root,
                        raw=section_name,
                        rules=tuple(rules),
                    ))
                section_name = line
                rules = []

        if rules:
            sections.append(Section(
                root=self.root,
                raw=section_name,
                rules=tuple(rules),
            ))
        return tuple(sections)

    @property
    def rules(self) -> Iterator[Rule]:
        for section in self.sections:
            yield from section.rules

    def find_rule(self, path: Path) -> Rule | None:
        for section in reversed(self.sections):
            rule = section.find_rule(path)
            if rule is not None:
                return rule
        return None



from __future__ import annotations

from argparse import ArgumentParser
from pathlib import Path

from ._base import Command


class OwnersOf(Command):
    """Show owners of the given file.
    """

    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        Command.init_parser(parser)
        parser.add_argument('paths', nargs='+')

    def run(self) -> int:
        paths = [Path(p).absolute() for p in self.args.paths]
        found = False
        for rule in self.code_owners.rules:
            if not any(rule.includes(p) for p in paths):
                continue
            for owner in rule.owners:
                self.print(owner)
                found = True
        return int(not found)



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from ._rule import Rule


@dataclass(frozen=True)
class Section:
    root: Path
    raw: str
    rules: tuple[Rule, ...]

    @cached_property
    def name(self) -> str | None:
        raw = self.raw.strip()
        without_comments = raw.split('#')[0]
        if not raw:
            return None
        return without_comments.lstrip('^[').rstrip(']')

    @cached_property
    def required(self) -> bool:
        return not self.raw.lstrip().startswith('^')

    def find_rule(self, path: Path) -> Rule | None:
        path = path.absolute()
        for rule in reversed(self.rules):
            if path in rule.paths:
                return rule
        for rule in reversed(self.rules):
            if rule.includes(path):
                return rule
        return None



from __future__ import annotations

from types import MappingProxyType

from ._base import Command
from ._owned_by import OwnedBy
from ._owners_of import OwnersOf
from ._tree import Tree
from ._version import Version


commands: MappingProxyType[str, type[Command]]
commands = MappingProxyType({
    'owned-by': OwnedBy,
    'owners-of': OwnersOf,
    'tree': Tree,
    'version': Version,
})

__all__ = [
    'commands',
    'Command',
]



from __future__ import annotations

from argparse import ArgumentParser
from dataclasses import dataclass
from enum import Enum
from functools import cached_property
from pathlib import Path
from typing import Callable

from ._base import Command


class Cov(Enum):
    FULL = 'full'  # the dir is 100% owned
    MOST = 'most'  # not the dir itself but everything inside is owned
    PART = 'part'  # some subdirs in the dir are owned
    NONE = 'none'  # nothing in the dir is owned


# Directories to skip by default
IGNORE = frozenset({'.git', '__pycache__'})


@dataclass
class PathInfo:
    # path to the analyzed dir
    path: Path

    # information about all subdirs, None if not analyzed (because the root is owned)
    children: list[PathInfo] | None

    # the owner team, empty string for not directly owned dirs
    owner: str

    @cached_property
    def cov(self) -> Cov:
        if self.children is None:
            return Cov.FULL
        if not self.children:
            return Cov.NONE
        if all(p.cov in (Cov.FULL, Cov.MOST) for p in self.children):
            return Cov.MOST
        if all(p.cov == Cov.NONE for p in self.children):
            return Cov.NONE
        return Cov.PART


class Tree(Command):
    """Show file tree with ownership info.
    """

    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        Command.init_parser(parser)

    def run(self) -> int:
        path_info = self._inspect(self.code_owners.root)
        self._show_tree(path_info, 0)
        return 0

    def _inspect(self, root: Path) -> PathInfo:
        rule = self.code_owners.find_rule(root)
        if rule is not None:
            return PathInfo(root, children=None, owner=rule.owners[0])

        path_infos: list[PathInfo] = []
        for path in sorted(root.iterdir()):
            if not path.is_dir():
                continue
            if path.name in IGNORE:
                continue
            path_infos.append(self._inspect(path))
        return PathInfo(root, children=path_infos, owner='')

    def _show_tree(self, info: PathInfo, level: int) -> None:
        if info.cov == Cov.FULL:
            self._show_path(self.colors.green, info, level)
            return
        if info.cov == Cov.MOST:
            self._show_path(self.colors.blue, info, level)
            return
        if info.cov == Cov.NONE:
            self._show_path(self.colors.red, info, level)
            return
        if info.cov == Cov.PART:
            assert info.children
            self._show_path(self.colors.yellow, info, level)
            for child in info.children:
                self._show_tree(child, level + 1)
            return
        raise RuntimeError('unreachable')

    def _show_path(self, color: Callable, info: PathInfo, level: int) -> None:
        width = level * 2 + len(info.path.name) + 1
        ident = 60 - width
        self.print(' |' * level, color(info.path.name), ' ' * ident + info.owner)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any


RED = '\033[31m'
GREEN = '\033[32m'
YELLOW = '\033[33m'
BLUE = '\033[94m'
MAGENTA = '\033[35m'
END = '\033[0m'


@dataclass
class Colors:
    disabled: bool

    def green(self, text: Any) -> str:
        if self.disabled:
            return str(text)
        return f'{GREEN}{text}{END}'

    def red(self, text: Any) -> str:
        if self.disabled:
            return str(text)
        return f'{RED}{text}{END}'

    def blue(self, text: Any) -> str:
        if self.disabled:
            return str(text)
        return f'{BLUE}{text}{END}'

    def magenta(self, text: Any) -> str:
        if self.disabled:
            return str(text)
        return f'{MAGENTA}{text}{END}'

    def yellow(self, text: Any) -> str:
        if self.disabled:
            return str(text)
        return f'{YELLOW}{text}{END}'



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from pathlib import Path


@dataclass(frozen=True)
class Rule:
    root: Path
    raw: str

    @cached_property
    def raw_path(self) -> str:
        return self.raw.split()[0]

    @cached_property
    def paths(self) -> tuple[Path, ...]:
        raw = self.raw_path.lstrip('/')
        if '*' in raw:
            paths = [path.absolute() for path in self.root.glob(raw)]
            paths_set = set(paths)
            # remove paths that are children of other paths
            deduplicated = []
            for path in paths:
                if not set(path.parents) & paths_set:
                    deduplicated.append(path)
            return tuple(deduplicated)

        path = Path(raw)
        if not path.exists():
            return tuple()
        return (path.absolute(),)

    @cached_property
    def owners(self) -> tuple[str, ...]:
        without_comments = self.raw.split('#')[0]
        return tuple(without_comments.split()[1:])

    def includes(self, path: Path) -> bool:
        """Check if the given path is included in the rule.
        """
        path = path.absolute()
        if path in self.paths:
            return True
        for parent in path.parents:
            if parent in self.paths:
                return True
        return False
